#ifndef LATTICE_H
#define LATTICE_H
#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include "StandardCollimator.h"
#include "FlukaCollimator.h"
#include "MagneticCollimator.h"
#include "CrystalCollimator.h"

#if defined(FLUKA)
#include "FlukaIO.h"
#endif

using namespace std;
/*
############################################################################################################

Class which represent the lattice of the experiment

###########################################################################################################
 */


class Lattice
{

public:


    //==========================Constructors, destructor======================

    Lattice() {};

    Lattice(Element start, const vector <int>& ips);

    Lattice(Element element1, Element element2, StandardCollimator stdcolli, FlukaCollimator flukacolli, MagneticCollimator magnetcolli, const vector <int>& ips, int size = 5, int npart = 1);

    virtual ~Lattice();

    //=============================== Other methods ===============================

    //method used to set the variable size_res

    void setsize_res(int s);


    //non-linear tracking of the particles through the accelerator

    void trackensemblechrom(vector <Particle>& bunch, const int& irev, const int& i0, const int& im, const double& Apr, const double& Zpr, const double& wecolli, const double& betgam, const int& nonlinflag, const int& scaleorbit, double& attr1, const int& idpart, const int& idelt, const int& outcoord, const string& plotflag, vector <vector <double> >& xco, vector <vector <double> >& yco, string outputpath, int RFflag, int& indication);


    //used to add a collimator to resColli, the lattice containing only the collimators

    void addCollimator(Collimator* colli);


    //used to add an element to the accelerator's lattice

    void addElement(Element* elt);


    //used to compute the 6th coordinate of the particles

    double time(Particle& p, const double& l, const double& betgam, const int& elt);


    //linear tracking of the particles; here, we only look for collision with primary collimators

    void trackensemblelinearnew(vector <Particle>& bunch, vector <Particle>& bunchhit, const int& nrev, const double& blowup2, const int& blowupperiod);


    //output of the x-, y- and t- coordinates of the particles in the file coordinates.dat; mostly used for tests; the choice of the particle is made in lattice.cc with the static int choicePart
    //the first column in the file coordinates.dat is the x-coordinate and the second one is the y-coordinate

    void outCoord(const Particle& p, const int& indic, const string& fileOut);



    //displays the x-, y- and t-coordinates of a chosen particle at each turn at a given element in the lattice; the result is written in the file coordinates_punctual.dat;
    //the chosen element number must be set in lattice.cc with the variable eltOutNber; the chosen particle must be set in lattice.cc with the variable choicePart

    void outPunct(const int& elt, const Particle& p, const Particle& p2, const double& var, string outputpath);


    //displays the phase advance and the the amount of provided energy for the passage through RF-cavities in the file "valuestestrf.dat"; used for tests; if their are more than one cavity in the turns,
    //for the first turn only one cavity is taken into account; to use it, the call must be uncomment in the file "lattice.cc"

    void outrf(const double& x1, const double& x2, string outputpath);


    //displays coordinates of all the particles at a certain element of the lattice in the file "coordinates_elt.dat"; the element is chosen throught the global variable eltOutNber in lattice.cc

    void outElt(const int& elt, const Particle& p, string outputpath, int& indication);


    void read(vector <Particle>& bunch);


    //=========================================== Attributs ============================================

#if defined(FLUKA)
    flukaio_connection_t* conn; //connection with an external server (e.g. Fluka)
#endif

    int size_res; //size of the lattice

    int npart; //number of particles
    vector <int> ips; //vector indicating the position of the collimators (gives the index of the collimators in the complete lattice, e.g. if the first collimator is the 5th element, we will have ips[0] = 5)

    vector <int> cocount;//count the number of particles crashing on each collimator

    vector < Element* > reseau; //lattice formed by all the elements
    vector < Collimator* > resColli; //vector containing only the primary collimators that are in the lattice

    //The following vectors contains the indices of the positions of such collimators

    vector <int> ii;//'TCI' collimators
    vector < int > ip;//'TCP', 'CRYSTAL' and 'TCRYO' collimators
    vector < int > is;//'TCS' collimators
    vector < int > it;//'TCT' collimators
    vector < int > itcla;//'TCLA' collimators

    vector < int > ipcoll; // vector giving the position like ips, but only for 'TCP' collimators

    vector <int> ixcormag;// vector of all i's such that  etype(i) starts with 'KICKER' of 'HKICKER' and px(i) != px(i-1)
    vector <int> iycormag;// same as above for for 'KICKER' OR 'VKICKER' and py
    vector <double> xcormag;// px(i)-px(i-1) for all the i's in ixcormag
    vector <double> ycormag;// same as above for py

    vector <double> hits;//positions of the hits
    vector <double> Aphit;//mass for particles hitting the aperture
    vector <double> Zphit;//charge for particles hitting the aperture
    vector <int> nhitcolli;//for each collimator, number of particles getting lost on it
    vector <int> nhitspoiler;//for each ion spoiler, number of particles getting lost on it

    vector <double> sh;//vector used to store the position of the elements during the computation (used in simulation.cc)

    //apertures (used in simulation.cc)
    vector <double> aperx1;
    vector <double> apery1;

    vector <double> apdepth;//a particle's perpendicular depth in a collimator as it hits the collimator. Measured at the beginning of the collimator
    vector <double> apdepth2;//a particle's perpendicular depth in a collimator as it hits the collimator. Measured at the end of the collimator

    //elements of the matrix R
    vector <double> R11X;
    vector <double> R12X;
    vector <double> R21X;
    vector <double> R22X;
    vector <double> R11Y;
    vector <double> R12Y;
    vector <double> R21Y;
    vector <double> R22Y;
    vector <double> Lelem;//distance between 2 consecutive elements
    //elements of the matrix RC
    vector <double> R11XC;
    vector <double> R12XC;
    vector <double> R21XC;
    vector <double> R22XC;
    vector <double> R11YC;
    vector <double> R12YC;
    vector <double> R21YC;
    vector <double> R22YC;

    double freqrf;//RF frequency
    double rfvoltage;//RF voltage
    double rfharmonic;//RF harmonic number
    double momentum;//momentum

    static int count;//count the number of particles lost on a crystal collimator
    static int turn;//count the number of turns in the ring
    static int choicePart;//identification of the particle chosen for the study in the file coordinates.dat
    static int eltOutNber;//id of the element for which we want the output in coordinates_punctual.dat

};


#endif

